from datetime import datetime
from pathlib import Path

import pytest
from approval_utilities.utilities.multiline_string_utils import remove_indentation_from
from pytest import ExitCode

from factories import make_raster_at, make_zarr_at, make_nc_at


@pytest.fixture
def standard_approval_test_directory():
    return Path(__file__).parent


@pytest.fixture
def make_tmp_approval_tif(standard_approval_test_directory):
    creates_files = []
    try:
        def _fn(values, name):
            file = make_raster_at(values, standard_approval_test_directory / name)
            creates_files.append(file)

        yield _fn
    finally:
        for f in creates_files:
            f.unlink()


def test_approvaltests_geo_data_settings(testdir, tmp_path):
    approval_root = tmp_path / "path/to/my/approved/geo/data/root/"
    approval_root.mkdir(parents=True, exist_ok=True)
    approval_input = approval_root / "subdir/to/input/data/"
    approval_input.mkdir(parents=True, exist_ok=True)
    approval_approved = approval_root / "subdir/to/approved/data/"
    approval_approved.mkdir(parents=True, exist_ok=True)
    testdir.makeini(f"""
        [pytest]
        approvaltests_geo_data_root = {approval_root}
        approvaltests_geo_input = subdir/to/input/data/
        approvaltests_geo_approved = subdir/to/approved/data/
    """)

    testdir.makepyfile(f"""
        import pytest
        def test_approved_geo_paths(approval_test_geo_data_root, approval_geo_input_directory, approved_geo_directory):
            assert approval_test_geo_data_root.as_posix() == '{approval_root.as_posix()}'
            assert approval_geo_input_directory.as_posix() == '{approval_input.as_posix()}'
            assert approved_geo_directory.as_posix() == '{approval_approved.as_posix()}'
    """)

    result = testdir.runpytest('-v')

    result.stdout.fnmatch_lines([
        '*::test_approved_geo_paths PASSED*',
    ])

    assert result.ret == 0


def test_approval_test_geo_data_root_option(testdir, tmp_path):
    custom_approval_root = tmp_path / "custom/path/to/my/approved/geo/data/root/"
    custom_approval_root.mkdir(parents=True, exist_ok=True)

    testdir.makeini("""
        [pytest]
        approvaltests_geo_data_root = /default/path/to/geo/test/data/root/
        approvaltests_geo_input = subdir/to/input/data/
        approvaltests_geo_approved = subdir/to/approved/data/
    """)

    testdir.makepyfile(f"""
        def test_custom_approval_test_geo_data_root(approval_test_geo_data_root):
            assert approval_test_geo_data_root.as_posix() == "{custom_approval_root.as_posix()}"
    """)

    result = testdir.runpytest(
        f'--approval-test-geo-data-root={custom_approval_root.as_posix()}',
        '-v'
    )

    result.stdout.fnmatch_lines([
        '*::test_custom_approval_test_geo_data_root PASSED*',
    ])

    assert result.ret == 0


def test_skip_tests_which_request_approval_root_but_it_does_not_exist(testdir, tmp_path):
    testdir.makeini("""
        [pytest]
        approvaltests_geo_data_root = /nonexistent/path/to/geo/test/data/root/
        approvaltests_geo_input = subdir/to/input/data/
        approvaltests_geo_approved = subdir/to/approved/data/
    """)

    testdir.makepyfile(f"""
        def test_skipped_because_of_missing_approval_root(approval_geo_input_directory):
            assert False
    """)

    result = testdir.runpytest('-v')
    result.stdout.fnmatch_lines([
        '*::test_skipped_because_of_missing_approval_root SKIPPED*',
    ])

    assert result.ret == 0


def test_verify_geo_tif(testdir, tmp_path):
    make_standard_geo_data_setting(testdir, tmp_path)

    tif_file = make_raster_at([[42]], tmp_path / "a_tif_to_test.tif",
                              dict(some=datetime(2022, 1, 1).strftime("%Y-%m-%d %H:%M:%S")))

    testdir.makepyfile(f"""
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from pytest_approvaltests_geo.scrubbers import make_scrubber_recurse
            from approvaltests.scrubbers import scrub_all_dates
            def test_verify_geo_tif(verify_geo_tif):
                verify_geo_tif("{tif_file.as_posix()}", 
                    options=GeoOptions().with_tags_scrubber(make_scrubber_recurse(scrub_all_dates)))
        """)

    result = testdir.runpytest('-v')

    result.stdout.fnmatch_lines([
        '+    "some": "<date0>"',
    ])

    assert result.ret == ExitCode.TESTS_FAILED


def make_standard_geo_data_setting(testdir, tmp_path):
    root = tmp_path / "root"
    input_data = root / "input"
    approved = root / "approved"
    input_data.mkdir(parents=True, exist_ok=True)
    approved.mkdir(parents=True, exist_ok=True)
    testdir.makeini(f"""
        [pytest]
        approvaltests_geo_data_root = {root.as_posix()}
        approvaltests_geo_input = input
        approvaltests_geo_approved = approved
    """)

    return root, input_data, approved


def test_verify_multiple_geo_tiffs(testdir, tmp_path):
    make_standard_geo_data_setting(testdir, tmp_path)

    tif_a = make_raster_at([[42]], tmp_path / "a_tif_to_test.tif", dict(band='VV'))
    tif_b = make_raster_at([[42]], tmp_path / "another_tif_to_test.tif", dict(band='VH'))

    testdir.makepyfile(f"""
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from approval_utilities.utilities.exceptions.exception_collector import gather_all_exceptions_and_throw
            def test_verify_multiple_geo_tiffs(verify_geo_tif):
                gather_all_exceptions_and_throw(["{tif_a.as_posix()}", "{tif_b.as_posix()}"], 
                    lambda tif: verify_geo_tif(tif, options=GeoOptions().with_scenario_by_tags(lambda t: (t['band'],))))
        """)

    result = testdir.runpytest('-v')

    result.stdout.fnmatch_lines([
        "*test_verify_multiple_geo_tiffs.VV.received.tif*",
        "*test_verify_multiple_geo_tiffs.VH.received.tif*",
    ])

    assert result.ret == ExitCode.TESTS_FAILED


def test_verify_raster_as_geo_tif(testdir, tmp_path):
    _, _, approved_dir = make_standard_geo_data_setting(testdir, tmp_path)

    make_raster_at([[1.0]],
                   approved_dir / "test_approvaltests_geo_extensions.test_verify_raster_as_geo_tif.approved.tif")
    testdir.makepyfile(f"""
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from pytest_approvaltests_geo.factories import make_raster
            def test_verify_raster_as_geo_tif(verify_raster_as_geo_tif):
                verify_raster_as_geo_tif(make_raster([[1.0]]))
        """)

    result = testdir.runpytest(Path(testdir.tmpdir), '-v')
    assert result.ret == ExitCode.OK


def test_verify_raster_as_geo_tif_with_tolerance(testdir, tmp_path):
    _, _, approved_dir = make_standard_geo_data_setting(testdir, tmp_path)

    make_raster_at([[1.1]],
                   approved_dir / "test_approvaltests_geo_extensions.test_verify_raster_as_geo_tif_with_tolerance.approved.tif")
    testdir.makepyfile(f"""
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from pytest_approvaltests_geo.factories import make_raster
            def test_verify_raster_as_geo_tif(verify_raster_as_geo_tif):
                verify_raster_as_geo_tif(make_raster([[1.0]]), options=GeoOptions()\\
                    .with_tolerance(rel_tol=0.05, abs_tol=0.051))
        """)

    result = testdir.runpytest(Path(testdir.tmpdir), '-v')
    assert result.ret == ExitCode.OK


def test_verify_raster_as_geo_tif_with_custom_tif_writer(testdir, tmp_path):
    _, _, approved_dir = make_standard_geo_data_setting(testdir, tmp_path)
    make_raster_at([[2.0]],
                   approved_dir / "test_approvaltests_geo_extensions.test_verify_raster_as_geo_tif_with_custom_tif_writer.approved.tif")

    testdir.makepyfile(f"""
            import rioxarray
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from pytest_approvaltests_geo.factories import make_raster
            def test_verify_raster_as_geo_tif(verify_raster_as_geo_tif):
                verify_raster_as_geo_tif(make_raster([[1.0]]), options=GeoOptions()\\
                    .with_tif_writer(lambda f, a: make_raster([[2.0]]).rio.to_raster(f)))
        """)

    result = testdir.runpytest(Path(testdir.tmpdir), '-v')
    assert result.ret == ExitCode.OK


def test_verify_multiple_rasters_as_geo_tif(testdir, make_tmp_approval_tif):
    make_tmp_approval_tif([[1]],
                          "test_approvaltests_geo_extensions.test_verify_multiple_rasters_as_geo_tif.0.approved.tif")
    make_tmp_approval_tif([[2]],
                          "test_approvaltests_geo_extensions.test_verify_multiple_rasters_as_geo_tif.1.approved.tif")
    testdir.makepyfile(f"""
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from pytest_approvaltests_geo.factories import make_raster
            from approvaltests.namer.default_namer_factory import NamerFactory
            from approval_utilities.utilities.exceptions.exception_collector import gather_all_exceptions_and_throw
            def test_verify_multiple_rasters_as_geo_tif(verify_raster_as_geo_tif):
                rasters = [make_raster([[1]]), make_raster([[2]])]
                gather_all_exceptions_and_throw([0, 1], lambda i: verify_raster_as_geo_tif(
                    rasters[i],
                    options=GeoOptions.from_options(NamerFactory.with_parameters(i))
                ))
        """)

    result = testdir.runpytest(Path(testdir.tmpdir), '-v')

    assert result.ret == ExitCode.OK


def test_verify_geo_zarr(testdir, tmp_path):
    _, _, approved_dir = make_standard_geo_data_setting(testdir, tmp_path)

    zarr_file = make_zarr_at([[1.0]], tmp_path / "a_zarr_to_test.zarr",
                             dict(some=datetime(2022, 1, 1).strftime("%Y-%m-%d %H:%M:%S")))
    make_zarr_at([[1.1]], approved_dir / "test_approvaltests_geo_extensions.test_verify_geo_zarr.approved.zarr",
                 dict(some=datetime(2022, 1, 2).strftime("%Y-%m-%d %H:%M:%S")))
    testdir.makepyfile(f"""
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from pytest_approvaltests_geo.scrubbers import make_scrubber_recurse
            from approvaltests.scrubbers import scrub_all_dates
            def test_verify_geo_zarr(verify_geo_zarr):
                verify_geo_zarr("{zarr_file.as_posix()}", options=GeoOptions()\\
                    .with_tags_scrubber(make_scrubber_recurse(scrub_all_dates))
                    .with_tolerance(rel_tol=0.05, abs_tol=0.051))
        """)

    result = testdir.runpytest(Path(testdir.tmpdir), '-v')
    assert result.ret == ExitCode.OK


def test_verify_parametrized_geo_zarr(testdir, tmp_path):
    _, _, approved_dir = make_standard_geo_data_setting(testdir, tmp_path)

    zarr_file = make_zarr_at([[1.0]], tmp_path / "a_zarr_to_test.zarr")
    make_zarr_at([[1.0]], approved_dir /
                 "test_approvaltests_geo_extensions.test_verify_parametrized_geo_zarr.scenario0.approved.zarr")
    testdir.makepyfile(f"""
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from approvaltests import Options
            def test_verify_geo_zarr(verify_geo_zarr, name_geo_scenario):
                verify_geo_zarr("{zarr_file.as_posix()}", options=GeoOptions\\
                    .from_options(Options().with_namer(name_geo_scenario("scenario0"))))
        """)

    result = testdir.runpytest(Path(testdir.tmpdir), '-v')
    assert result.ret == ExitCode.OK


def test_verify_geo_zarr_without_options(testdir, tmp_path):
    _, _, approved_dir = make_standard_geo_data_setting(testdir, tmp_path)

    zarr_file = make_zarr_at([[1.0]], tmp_path / "a_zarr_to_test.zarr")
    make_zarr_at([[1.0]], approved_dir /
                 "test_approvaltests_geo_extensions.test_verify_geo_zarr_without_options.approved.zarr")
    testdir.makepyfile(f"""
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from approvaltests import Options
            def test_verify_geo_zarr(verify_geo_zarr, name_geo_scenario):
                verify_geo_zarr("{zarr_file.as_posix()}")
        """)

    result = testdir.runpytest(Path(testdir.tmpdir), '-v')
    assert result.ret == ExitCode.OK


def test_verify_geo_nc(testdir, tmp_path):
    _, _, approved_dir = make_standard_geo_data_setting(testdir, tmp_path)

    nc_file = make_nc_at([[1.0]], tmp_path / "a_nc_to_test.nc",
                         dict(some=datetime(2022, 1, 1).strftime("%Y-%m-%d %H:%M:%S")))
    make_nc_at([[1.1]], approved_dir / "test_approvaltests_geo_extensions.test_verify_geo_nc.approved.nc",
               dict(some=datetime(2022, 1, 2).strftime("%Y-%m-%d %H:%M:%S")))
    testdir.makepyfile(f"""
            from pytest_approvaltests_geo.geo_options import GeoOptions
            from pytest_approvaltests_geo.scrubbers import make_scrubber_recurse
            from approvaltests.scrubbers import scrub_all_dates
            def test_verify_geo_nc(verify_geo_nc):
                verify_geo_nc("{nc_file.as_posix()}", options=GeoOptions()\\
                    .with_tags_scrubber(make_scrubber_recurse(scrub_all_dates))
                    .with_tolerance(rel_tol=0.05, abs_tol=0.051))
        """)

    result = testdir.runpytest(Path(testdir.tmpdir), '-v')
    assert result.ret == ExitCode.OK


def test_doc_tests_still_working(pytester, tmp_path):
    make_standard_geo_data_setting(pytester, tmp_path)

    module = pytester.mkpydir("some_module")
    (module / "a_class_with_doc.py").write_text(remove_indentation_from(f'''
            class AClassWithDoc:
                """
                >>> AClassWithDoc(42).the_answer
                42
                """
                def __init__(self, answer):
                    self.the_answer = answer
            '''))

    result = pytester.runpytest(Path(pytester.path), '--doctest-modules')
    assert result.ret == ExitCode.OK



from typing import Callable, Dict, Union, List, Tuple, get_args, Any

from approvaltests.scrubbers.scrubbers import Scrubber

JsonLikeCollection = Union[Dict, List, Tuple]
RecursiveScrubber = Callable[[JsonLikeCollection], JsonLikeCollection]


def scrub_recursive(tags: JsonLikeCollection, scrubber: Scrubber) -> JsonLikeCollection:
    if isinstance(tags, Dict):
        return scrub_dict_recursive(tags, scrubber)
    if isinstance(tags, List):
        return scrub_list_recursive(tags, scrubber)
    if isinstance(tags, Tuple):
        return scrub_tuple_recursive(tags, scrubber)


def scrub_dict_recursive(tags: Dict, scrubber: Scrubber) -> Dict:
    scrubbed = {}
    for k, v in tags.items():
        if isinstance(k, str):
            k = scrubber(k)
        v = scrub_element(v, scrubber)
        scrubbed[k] = v
    return scrubbed


def scrub_element(element: Any, scrubber: Scrubber) -> Any:
    if isinstance(element, str):
        element = scrubber(element)
    if isinstance(element, get_args(JsonLikeCollection)):
        element = scrub_recursive(element, scrubber)
    return element


def scrub_list_recursive(tags: List, scrubber: Scrubber) -> List:
    return [scrub_element(t, scrubber) for t in tags]


def scrub_tuple_recursive(tags: Tuple, scrubber: Scrubber) -> Tuple:
    return tuple(scrub_element(t, scrubber) for t in tags)


def make_scrubber_recurse(scrubber: Scrubber) -> RecursiveScrubber:
    return lambda elems: scrub_recursive(elems, scrubber)


def identity_recursive_scrubber(tags: Dict) -> Dict:
    return tags


def scrub_xarray_data(a, scrubber):
    a.attrs = scrubber(a.attrs)
    for name in a.coords:
        a[name].attrs = scrubber(a[name].attrs)
    for name in getattr(a, 'data_vars', {}):
        a[name].attrs = scrubber(a[name].attrs)

    return a



from approvaltests import StackFrameNamer


class StackFrameNamerWithExternalDataDir(StackFrameNamer):
    def __init__(self, data_root):
        super().__init__()
        self._data_root = data_root

    def get_directory(self) -> str:
        return self._data_root



import os
from difflib import unified_diff
from pathlib import Path
from typing import Sequence, Optional

import xarray as xr
from approval_utilities.utils import to_json

from pytest_approvaltests_geo.differs.difference import DiffType, Difference, calculate_pixel_diff_stats, \
    add_common_meta_data_diffs
from pytest_approvaltests_geo.float_utils import Tolerance
from pytest_approvaltests_geo.geo_io import read_array_and_tags
from pytest_approvaltests_geo.scrubbers import RecursiveScrubber, identity_recursive_scrubber, scrub_xarray_data


class DifferOfGeoTiffs:
    def __init__(self, recursive_scrubber: Optional[RecursiveScrubber] = None,
                 float_tolerance: Optional[Tolerance] = None):
        self._recursive_scrubber = recursive_scrubber or identity_recursive_scrubber
        self._float_tolerance = float_tolerance or Tolerance()

    def diffs(self, received_path: Path, approved_path: Path) -> Sequence[Difference]:
        diffs = []
        with read_array_and_tags(received_path) as (received_pixels, received_tags), \
                read_array_and_tags(approved_path) as (approved_pixels, approved_tags):
            diff_tags = self._calculate_tags_diff(approved_path, approved_tags, received_path, received_tags)
            if diff_tags:
                diffs.append(Difference(diff_tags, DiffType.TAGS))

            received_pixels = scrub_xarray_data(received_pixels, self._recursive_scrubber)
            approved_pixels = scrub_xarray_data(approved_pixels, self._recursive_scrubber)
            diffs = add_common_meta_data_diffs(received_pixels, approved_pixels, diffs)

            try:
                xr.testing.assert_allclose(received_pixels, approved_pixels, **self._float_tolerance.to_kwargs())
            except AssertionError as assertion_diff:
                diff_px_stats = calculate_pixel_diff_stats(approved_pixels, received_pixels)
                diffs.append(Difference(f"pixel differences statistics:\n{str(diff_px_stats)}", DiffType.PIXEL_STATS))
                diffs.append(Difference(str(assertion_diff), DiffType.DATASET))
        return diffs

    def _calculate_tags_diff(self, approved_path, approved_tags, received_path, received_tags):
        approved_text = f"{to_json(self._recursive_scrubber(approved_tags))}\n"
        received_text = f"{to_json(self._recursive_scrubber(received_tags))}\n"
        return "\n".join(unified_diff(
            approved_text.splitlines(),
            received_text.splitlines(),
            os.path.basename(approved_path),
            os.path.basename(received_path)
        )).strip()



from datetime import datetime

import numpy as np
import pytest
from approvaltests.scrubbers import create_regex_scrubber

from factories import make_raster_at
from pytest_approvaltests_geo.reporters.report_geo_tiffs import ReportGeoTiffs
from pytest_approvaltests_geo.scrubbers import make_scrubber_recurse

PIXEL_DIFF_REPORT_PATTERN = "[data][band=1"


@pytest.fixture
def reporter():
    return ReportGeoTiffs()


def test_identical_geo_tiffs_report_nothing(reporter, tmp_path, capsys):
    geo_tif = make_raster_at([[42]], tmp_path / "geo_tif.tif", dict(some='tag'))
    assert reporter.report(geo_tif.as_posix(), geo_tif.as_posix())
    assert capsys.readouterr().out == ""


def test_report_pixel_difference_statistics(reporter, tmp_path, capsys):
    received = make_raster_at([[0, 0], [0, 0]], tmp_path / "received.tif", dict(some='tag'))
    approved = make_raster_at([[2, 4], [-2, 1]], tmp_path / "approved.tif", dict(some='tag'))
    assert reporter.report(received.as_posix(), approved.as_posix())
    assert "min=1, max=4, mean=2.25, median=2" in capsys.readouterr().out


def test_report_meta_data_differences(reporter, tmp_path, capsys):
    received = make_raster_at([[42]], tmp_path / "received.tif", dict(some='tag'))
    approved = make_raster_at([[42]], tmp_path / "approved.tif", dict(some='other'))
    assert reporter.report(received.as_posix(), approved.as_posix())
    output = capsys.readouterr().out
    assert '+    "some": "tag"' in output and '-    "some": "other"' in output


def test_report_scrubbed_data(tmp_path, capsys):
    date_scrubber = create_regex_scrubber(
        r"\d{4}-\d{2}-\d{2}T\d{2}-\d{2}-\d{2}", lambda t: f"<date{t}>"
    )
    scrubbing_reporter = ReportGeoTiffs(make_scrubber_recurse(date_scrubber))
    received = make_raster_at([[0, 0], [0, 0]], tmp_path / "received.tif",
                              {"some": datetime(2022, 1, 1).strftime("%Y-%m-%dT%H-%M-%S"),
                               datetime(2022, 2, 1).strftime("%Y-%m-%dT%H-%M-%S"): 42})
    approved = make_raster_at([[2, 4], [-2, 1]], tmp_path / "approved.tif",
                              {"other": datetime(2022, 1, 1).strftime("%Y-%m-%dT%H-%M-%S"),
                               datetime(2022, 2, 1).strftime("%Y-%m-%dT%H-%M-%S"): 21})
    assert scrubbing_reporter.report(received.as_posix(), approved.as_posix())
    output = capsys.readouterr().out
    assert '+    "some": "<date0>"' in output and '-    "other": "<date0>"' in output
    assert '+    "<date0>": "42"' in output and '-    "<date0>": "21"' in output


def test_report_has_maximum_length_when_reporting_differing_pixels(reporter, tmp_path, capsys):
    received = make_raster_at(np.zeros((100, 100)), tmp_path / "received.tif", dict(some='tag'))
    approved = make_raster_at(np.ones((100, 100)), tmp_path / "approved.tif", dict(some='tag'))
    assert reporter.report(received.as_posix(), approved.as_posix())
    diff_pixel_lines = [line for line in capsys.readouterr().out.splitlines() if PIXEL_DIFF_REPORT_PATTERN in line]
    assert len(diff_pixel_lines) < 10
